@@ -105,6 +105,7 @@ class go(commands.Command):
 		direction_help = [
 			direction_fmt.format(to=cls.app.GO_DETAILS, help=_("Open details")),
 			direction_fmt.format(to=cls.app.GO_TAG, help=_("Open tag")),
+			direction_fmt.format(to=cls.app.GO_TODO, help=_("Open list of TODO flags")),
 			direction_fmt.format(to=cls.app.GO_LOG, help=_("Return to the first view")),
 			direction_fmt.format(to=cls.app.GO_LEFT, help=_("Go to the previous view")),
 			direction_fmt.format(to=cls.app.GO_RIGHT, help=_("Go to the next view")),
@@ -710,7 +711,7 @@ class set(commands.Command):
 	@classmethod
 	def add_internal_functions_to_help(cls, out, view, model):
 		out.append("")
-		out.append(_("internal functions which can be used in log view:").format(view=view))
+		out.append(_("internal functions which can be used in {view}:").format(view=view))
 		for funcname in sorted(model.command_functions.keys()):
 			func = model.command_functions[funcname]
 			funcname = model.INTERNAL_FUNCTION_PREFIX + funcname
