@@ -38,8 +38,9 @@ def run():
 	kw = main.parse_common_args(args, doc=__doc__)
 
 	diff_args = args
+	ignore_repo = '--no-index' in diff_args
 
-	a = main.App(diff=diff_args, **kw)
+	a = main.App(diff=diff_args, ignore_repo=ignore_repo, **kw)
 	while a.continue_running:
 		a.run()
 		a.run_external_if_requested()
