@@ -9,7 +9,7 @@ import shutil
 import gettext
 _ = gettext.gettext
 
-from .api_subprocess import Runner, CommandError
+from .api_subprocess import Runner, CommandError, TimeoutExpired
 from .string_formatter import formatter
 from .color_decoder import ColorDecoder
 from . import utils
@@ -32,7 +32,8 @@ title_unstaged = _('Local uncommitted changes, not checked in to index')
 title_unstaged_and_untracked = formatter.format(_('{title_unstaged} and {title_untracked:l}'), title_unstaged=title_unstaged, title_untracked=title_untracked)
 title_staged   = _('Local changes checked in to index but not committed')
 title_stashes_group = _('{number} stashes')
-title_todo = _('{number} TODO flags')
+title_todo_log = _('TODO flags')
+title_todo_details = _('{number} TODO flags')
 title_referencedby = _('This commit is referenced by the following commits:')
 title_commit = _("commit {hash_id}{refnames}")
 pattern_refnames = _(" ({refnames})")
@@ -40,6 +41,8 @@ refnames_sep = _(", ")
 
 error_no_unreachable_commits = _('no unreachable commits')
 
+hint_timeout = _("Either speed up the command (e.g. in case of a grep ignoring untracked files can significantly reduce the amount of files to be searched - depending on the repository and the working directory) or change the timeout in the config file with `set log-view.timeout=<s>` or `set details-view.timeout=<s>` where <s> is the time in seconds or -1 to disable the timeout.")
+
 
 # ---------- model classes ----------
 
@@ -139,6 +142,23 @@ class Model(Runner):
 
 		return [[value if x == wildcard else x for x in subcmd] for subcmd in cmd]
 
+	@classmethod
+	def replace_in_cmd_with_several(cls, cmd, wildcard, value):
+		if not cls.are_several_commands(cmd):
+			cmd = [cmd]
+
+		out = []
+		for subcmd in cmd:
+			try:
+				i = subcmd.index(wildcard)
+			except ValueError:
+				pass
+			else:
+				subcmd = subcmd[:i] + value + subcmd[i+1:]
+			out.append(subcmd)
+
+		return out
+
 	@classmethod
 	def replace_characters_in_cmd(cls, cmd, old, new, count=-1):
 		"""
@@ -254,6 +274,7 @@ class Model(Runner):
 		cls._remote_branches = None
 
 
+_cmd_grep_todo = ["git", "grep", "--no-index", "--exclude-standard", "-n", "-I", "-E", '(#|%|"|//)TODO\>|\<TODO[:]', '--', settings.WC_GIT_ROOT]
 
 class LogModel(Model):
 
@@ -263,6 +284,9 @@ class LogModel(Model):
 	cmd_log = [["$stashes"], ["$todo"], ["$status"], ["git", "log", "--color", "--graph", "--date=relative", FLAG_CUSTOM_FORMAT, "--abbrev-commit"]]
 	cmd_stash_list = ["git", "stash", "list", "--color", "--pretty=format:%C(cyan)* %Cred%gd%Creset - ^%s %Cgreen(%cr)%Creset", "--abbrev-commit"]
 	cmd_unreachable = ["git", "fsck", "--unreachable"]
+	cmd_grep_todo_quiet = Model.replace_in_cmd_with_several(_cmd_grep_todo, 'grep', ['grep', '--quiet'])
+
+	timeout = .5
 
 	@classmethod
 	def append_cmd_log(cls, options):
@@ -314,6 +338,7 @@ class LogModel(Model):
 
 		self.ln_pattern = cls.ln_pattern.replace(align_char, "", 1)
 		self.ln_pattern_error = cls.ln_pattern_error.replace(align_char, "", 1)
+		self.ln_pattern_error_hint = cls.ln_pattern_error_hint.replace(align_char, "", 1)
 
 	def restore_align_character(self):
 		cls = type(self)
@@ -325,10 +350,12 @@ class LogModel(Model):
 
 		self.ln_pattern = cls.ln_pattern
 		self.ln_pattern_error = cls.ln_pattern_error
+		self.ln_pattern_error_hint = cls.ln_pattern_error_hint
 
 	WC_LOGENTRY = '{logentry}'
 	ln_pattern = '* ^<color=default>' + WC_LOGENTRY
-	ln_pattern_error = '\n<color=%color.echo.error%>{err}</color>\nsee %{setting}%\ncmd = {cmd}\n'
+	ln_pattern_error = '\n<color=%color.echo.error%>{err}</color>\nsee %{setting}%\ncmd = {cmd}\n{hint}'
+	ln_pattern_error_hint = '\n{hint}\n\n'
 
 	show_untracked = True
 	untracked_files_as_separate_group = False
@@ -337,6 +364,9 @@ class LogModel(Model):
 	stashed_changes_reversed_order = False
 	stashed_changes_group = None  # True|False|None. None = auto
 
+	show_list_of_todo_flags = None  # True|False|None. None = auto
+	show_list_of_todo_flags_on_timeout = False
+
 	re_hex = r'[0-9A-F]'
 	reo_hash_id = re.compile(r"(?:\b|(?<=m))(%(re_hex)s{7,40})\b" % dict(re_hex=re_hex), re.IGNORECASE)
 	reo_stash_id = re.compile(r"stash@\{\d+\}", re.IGNORECASE)
@@ -357,7 +387,11 @@ class LogModel(Model):
 		return log
 
 	def format_command_error(self, e, setting):
-		ln = self.ln_pattern_error.format(err=e.err, cmd=e.cmd, setting=setting)
+		if isinstance(e.exception, TimeoutExpired):
+			hint = self.ln_pattern_error_hint.format(hint=hint_timeout)
+		else:
+			hint = ''
+		ln = self.ln_pattern_error.format(err=e.err, cmd=e.cmd, setting=setting, hint=hint)
 		ln = utils.colored_str_to_markup(ln, self.app.define_color)
 		return [(ln, None, TYPE_ERROR)]
 
@@ -464,29 +498,38 @@ class LogModel(Model):
 			cmd = ['$todo'] + args
 			raise CommandError(cmd, "todo takes no arguments")
 
+		out_true = [(self.format_log_entry(settings.COLOR_LOG_TODO, title_todo_log), ID_TODO, TYPE_TODO)]
+		if self.show_list_of_todo_flags:
+			return out_true
+		elif self.show_list_of_todo_flags is False:
+			return []
+
 		root = get_git_root()
-		self.cmd = self.replace_in_cmd(DetailsModel.cmd_grep_todo, settings.WC_GIT_ROOT, root)
+		self.cmd = self.replace_in_cmd(self.cmd_grep_todo_quiet, settings.WC_GIT_ROOT, root)
 		try:
 			lines = super().get_lines()
 		except CommandError as e:
+			if isinstance(e.exception, TimeoutExpired):
+				if e.lines or self.show_list_of_todo_flags_on_timeout:
+					# Tests indicate that e.exception.stdout => e.lines is always None. So this is never called.
+					return out_true
+				else:
+					return []
 			if e.returncode == 1:
 				# no todo flags found
 				return []
 			return self.format_command_error(e, setting='details-view.cmd.todo.grep')
 		else:
-			if not lines:
-				return []
-
-			ln = formatter.format(title_todo, number=len(lines))
-			ln = self.format_log_entry(settings.COLOR_LOG_TODO, ln)
-			return [(ln, ID_TODO, TYPE_TODO)]
+			return out_true
 
 	todo.arguments = ''
 	todo.description = _('Show TODO flags')
 	todo.settings = [
-		'details-view.cmd.todo.grep',
+		'log-view.cmd.todo.grep',
+		'log-view.show-todo-flags',
+		'log-view.show-todo-flags-on-timeout',
 		'log-view.line-pattern',
-		'title.todo',
+		'title.todo.log-view',
 		settings.COLOR_LOG_TODO,
 	]
 
@@ -550,15 +593,13 @@ class UnreachableModel(LogModel):
 		try:
 			lines = Model.get_lines(self)
 		except CommandError as e:
-			ln = self.ln_pattern_error.format(err=e.err, cmd=e.cmd)
-			ln = utils.colored_str_to_markup(ln, self.app.define_color)
-			return [(ln, None, TYPE_ERROR)]
+			return self.format_command_error(e, setting='log-view.cmd.unreachable')
 
 		lines = [ln.split() for ln in lines]
 		lines = [ln[2] for ln in lines if ln[0] == "unreachable" and ln[1] == "commit"]
 		if not lines:
 			err = error_no_unreachable_commits + "\n"
-			ln = self.ln_pattern_error.format(err=err, cmd=self.cmd)
+			ln = self.ln_pattern_error.format(err=err, cmd=self.cmd, setting='log-view.cmd.unreachable', hint='')
 			ln = utils.colored_str_to_markup(ln, self.app.define_color)
 			return [(ln, None, TYPE_ERROR)]
 
@@ -592,13 +633,15 @@ class DetailsModel(Model):
 	cmd_staged = cmd_unstaged + ["--cached"]
 	cmd_untracked = ["$untracked"]
 	cmd_todo = ["$todo"]
-	cmd_grep_todo = ["git", "grep", "--no-index", "--exclude-standard", "-n", "-I", "-E", '(#|%|"|//)TODO\>|\<TODO[:]', '--', settings.WC_GIT_ROOT]
+	cmd_grep_todo = _cmd_grep_todo
 	cmd_pattern_stashed = [
 			["git", "log", "-g", "-1", "--pretty=format:%C(yellow)%gD%Creset%ncommit %H%nAuthor: %an <%ae>%nDate:   %ad%w(,4,4)%n%+B%n%n", "--color", WC_HASH_ID],
 			["git", "stash", "show", "--patch", "--color", WC_HASH_ID],
 	]
 	cmd_referencing_commits = [w for w in LogModel.cmd_log[-1] if w != "--graph"] + ["--branches"]
 
+	timeout = 2
+
 	log_format_options = (
 		'--no-decorate',
 		'--decorate',
@@ -858,7 +901,7 @@ class DetailsModel(Model):
 
 		return [([(settings.COLOR_DETAILS_UNTRACKED, utils.colored_str_to_markup(formatter.format(pattern, filename=fn), self.app.define_color))], TYPE_UNTRACKED, FileInfo(path(fn))) for fn in filenames]
 
-	def format_command_error(self, e, action=None, setting=None):
+	def format_command_error(self, e: CommandError, action: 'str|None' = None, setting: 'str|None' = None):
 		cmd = "cmd = %r\n" % (e.cmd,)
 		err = e.err
 		if action:
@@ -866,7 +909,12 @@ class DetailsModel(Model):
 		if setting:
 			err += "\n" + _("see %{setting}%").format(setting=setting)
 
-		return [cmd, (err, TYPE_ERROR)]
+		if isinstance(e.exception, TimeoutExpired):
+			hint = hint_timeout
+		else:
+			hint = ""
+
+		return [cmd, (err, TYPE_ERROR), hint]
 
 	def insert_linenumbers(self, out):
 		self.file_info = None
@@ -950,26 +998,15 @@ class DetailsModel(Model):
 		return "%" + str(width) + "s"
 
 	def insert_linenumbers__process_content_line(self, ns):
-		orig_ln = ns.out[ns.i]
-		ln = orig_ln
-
-		if self.insert_align_character:
-			ln = self.insert_align_char(ln, ns.i_align_character)
-
+		ln = ns.out[ns.i]
 		ns.line_info = LineInfo()
-		ln = (ln, TYPE_NUMBERED_LINE, self.indent_broken_code, ns.line_info)
-		ns.out[ns.i] = ln
-
-		self.increment_line_number_and_set_line_info(orig_ln, ns)
-
-	def increment_line_number_and_set_line_info(self, orig_ln, ns):
 		ns.line_info.linenumber_after = ns.linenumber_after
 		ns.line_info.linenumber_before_right = ns.linenumber_before_right
 		ns.line_info.linenumber_before_left = ns.linenumber_before_left
 		ns.line_info.is_modified = False
 		no_increment = True
 
-		if ns.reo_line_exists_after.match(orig_ln):
+		if ns.reo_line_exists_after.match(ln):
 			ns.line_info.formatted_linenumber_after = ns.fmt_linenumber_after % ns.linenumber_after
 			ns.linenumber_after += 1
 			no_increment = False
@@ -977,7 +1014,7 @@ class DetailsModel(Model):
 			ns.line_info.formatted_linenumber_after = ns.fmt_linenumber_after % ""
 			ns.line_info.is_modified = True
 
-		if ns.reo_line_existed_left.match(orig_ln):
+		if ns.reo_line_existed_left.match(ln):
 			ns.line_info.formatted_linenumber_before_left = ns.fmt_linenumber_before_left % ns.linenumber_before_left
 			ns.linenumber_before_left += 1
 			no_increment = False
@@ -985,7 +1022,7 @@ class DetailsModel(Model):
 			ns.line_info.formatted_linenumber_before_left = ns.fmt_linenumber_before_left % ""
 			ns.line_info.is_modified = True
 
-		if ns.reo_line_existed_right.match(orig_ln):
+		if ns.reo_line_existed_right.match(ln):
 			ns.line_info.formatted_linenumber_before_right = ns.fmt_linenumber_before_right % ns.linenumber_before_right
 			ns.linenumber_before_right += 1
 			no_increment = False
@@ -1005,10 +1042,17 @@ class DetailsModel(Model):
 			ns.linenumber_after += 1
 			ns.linenumber_before_left += 1
 			ns.linenumber_before_right += 1
+			if self.insert_align_character:
+				ln = ALIGN_CHAR + ln
+		elif self.insert_align_character:
+			ln = self.insert_align_char(ln, ns.i_align_character)
 
 		if ns.linenumber_after > ns.last_linenumber_after and ns.linenumber_before_right > ns.last_linenumber_before_right and ns.linenumber_before_left > ns.last_linenumber_before_left:
 			ns.look_for_hint = True
 
+		ln = (ln, TYPE_NUMBERED_LINE, self.indent_broken_code, ns.line_info)
+		ns.out[ns.i] = ln
+
 	@staticmethod
 	def insert_align_char(ln, pos_align_character):
 		n_escape = 0
@@ -1073,61 +1117,71 @@ class DetailsModel(Model):
 			cmd = ['$todo'] + args
 			raise CommandError(cmd, "todo takes no arguments")
 
+		timeout_expired = None
 		root = get_git_root()
 		self.cmd = self.replace_in_cmd(self.cmd_grep_todo, settings.WC_GIT_ROOT, root)
 		try:
 			lines = super().get_lines()
 		except CommandError as e:
-			if e.returncode == 1:
+			if isinstance(e.exception, TimeoutExpired) and e.lines:
+				lines = e.lines
+				timeout_expired = e
+			elif e.returncode == 1:
 				# no todo flags found
-				return [[(settings.COLOR_TITLE, title_todo.format(number=0))]]
+				lines = []
 			return self.format_command_error(e, setting='details-view.cmd.todo.grep')
-		else:
-			out = []
-			n = 0
-			for ln in lines:
-				ln = ln.replace('\t', self.app.DetailsView.TAB)
-				try:
-					filename, linenumber, todoflag = ln.split(':', 2)
-					linenumber = int(linenumber)
-				except ValueError:
-					out.append(ln)
+
+		out = []
+		n = 0
+		for ln in lines:
+			ln = ln.replace('\t', self.app.DetailsView.TAB)
+			try:
+				filename, linenumber, todoflag = ln.split(':', 2)
+				linenumber = int(linenumber)
+			except ValueError:
+				out.append(ln)
+			else:
+				filename = self.decode_filename(filename)
+				if filename.startswith('"') and filename.endswith('"'):
+					filename = filename[1:-1]
+				absfilename = os.path.abspath(filename)
+
+				if self.todo_relative == settings.RELATIVE_NAME_ONLY:
+					filename = os.path.split(filename)[1]
+				elif self.todo_relative == settings.RELATIVE_ROOT:
+					filename = absfilename
+				elif self.todo_relative == settings.RELATIVE_CWD:
+					pass
+				elif self.todo_relative == settings.RELATIVE_GIT:
+					filename = os.path.relpath(absfilename, root)
 				else:
-					filename = self.decode_filename(filename)
-					if filename.startswith('"') and filename.endswith('"'):
-						filename = filename[1:-1]
-					absfilename = os.path.abspath(filename)
+					assert False, 'invalid value for self.todo_relative: %r' % self.todo_relative
 
-					if self.todo_relative == settings.RELATIVE_NAME_ONLY:
-						filename = os.path.split(filename)[1]
-					elif self.todo_relative == settings.RELATIVE_ROOT:
-						filename = absfilename
-					elif self.todo_relative == settings.RELATIVE_CWD:
-						pass
-					elif self.todo_relative == settings.RELATIVE_GIT:
-						filename = os.path.relpath(absfilename, root)
-					else:
-						assert False, 'invalid value for self.todo_relative: %r' % self.todo_relative
+				if self.todo_strip_indentation:
+					todoflag = todoflag.strip()
+				if self.todo_only:
+					i = todoflag.index('TODO') + 4
+					todoflag = todoflag[i:].lstrip(':').lstrip()
 
-					if self.todo_strip_indentation:
-						todoflag = todoflag.strip()
-					if self.todo_only:
-						i = todoflag.index('TODO') + 4
-						todoflag = todoflag[i:].lstrip(':').lstrip()
+				markup = formatter.format(self.pattern_todo, filename=filename, linenumber=linenumber, todoflag=todoflag)
+				markup = utils.colored_str_to_markup(markup, self.app.define_color)
+				out.append((markup, TYPE_TODO, FileInfo(absfilename), LineInfo(linenumber)))
+				n += 1
 
-					markup = formatter.format(self.pattern_todo, filename=filename, linenumber=linenumber, todoflag=todoflag)
-					markup = utils.colored_str_to_markup(markup, self.app.define_color)
-					out.append((markup, TYPE_TODO, FileInfo(absfilename), LineInfo(linenumber)))
-					n += 1
-			out.insert(0, [(settings.COLOR_TITLE, title_todo.format(number=n))])
-			out.insert(1, '')
-			return out
+		out.insert(0, [(settings.COLOR_TITLE, title_todo_details.format(number=n))])
+		out.insert(1, '')
+
+		if timeout_expired:
+			out.extend(self.format_command_error(timeout_expired, setting='details-view.cmd.todo.grep'))
+
+		return out
 
 	todo.arguments = ''
 	todo.description = _('Show a list of the TODO flags in the code')
 	todo.settings = [
 		'details-view.cmd.todo.grep',
-		'details-view.todo.line-pattern'
+		'details-view.todo.line-pattern',
+		'title.todo.details-view',
 	]
 
 	def referencedby(self, args):
@@ -1505,7 +1559,7 @@ class Opener:
 		initattr(self, "editor_%s_linenumber" % name, linenumber)
 		initattr(self, "editor_%s_readonly" % name, readonly)
 		settings.settings["editor.%s.command" % name] = ("app.model.opener.editor_%s_command" % name, str, _("the value of the EDITOR environment variable which activates the corresponding editor settings"))
-		settings.settings["editor.%s.linenumber" % name] = ("app.model.opener.editor_%s_linenumber" % name, str, _("a command line argument which is added if a specific line shall be selected, the wild card {ln} specifies the line number"))
+		settings.settings["editor.%s.linenumber" % name] = ("app.model.opener.editor_%s_linenumber" % name, str, _("a command line argument which is added if a specific line shall be selected, the wildcard {ln} specifies the line number"))
 		settings.settings["editor.%s.readonly" % name] = ("app.model.opener.editor_%s_readonly" % name, str, _("a command line argument which is added if a file shall be opened read only"))
 
 	def get_editor(self):
@@ -1515,6 +1569,12 @@ class Opener:
 			if name is None:
 				# in case EDITOR is an absolute path
 				name = self._map.get(os.path.split(cmd)[1], None)
+			if name is None:
+				# in case EDITOR contains flags
+				name = self._map.get(shlex.split(cmd)[0], None)
+			if name is None:
+				# in case EDITOR contains flags and is an absolute path
+				name = self._map.get(os.path.split(shlex.split(cmd)[0])[1], None)
 		else:
 			name = self.get_default_editor_name()
 			cmd = self.get_editor_argument(name, "command")
@@ -1559,13 +1619,18 @@ class Opener:
 		if line_number == 0:
 			line_number = 1
 		runner = Runner()
-		editor_name, cmd = self.get_editor()
+		editor_name, editor_cmd = self.get_editor()
+		cmd = shlex.split(editor_cmd)
 		if line_number is not None:
-			cmd += " " + self.get_editor_argument(editor_name, "linenumber").format(ln=line_number)
+			arg = self.get_editor_argument(editor_name, "linenumber").format(ln=line_number)
+			if arg:
+				cmd.append(arg)
 		if read_only:
-			cmd += " " + self.get_editor_argument(editor_name, "readonly")
-		cmd += " " + shlex.quote(fn)
-		runner.run_interactive(cmd, shell=True)
+			arg = self.get_editor_argument(editor_name, "readonly")
+			if arg:
+				cmd.append(arg)
+		cmd.append(fn)
+		runner.run_interactive(cmd)
 
 	def check_editor(self):
 		editor_name, cmd = self.get_editor()
